# /packages/intranet-rest-fs-openacs/www/file-tree.json.tcl
#
# Copyright (C) 2013-2018 ]project-open[

ad_page_contract {
    Returns a JSON tree structure suitable for batch-loading a project TreeStore
    @param object_id ]po[ Business Object to which the FS belongs. 
           0 identifies the global "Home" FS
    @author frank.bergmann@project-open.com
    
    @param node Passed by ExtJS to load sub-trees of a tree.
                Normally not used, just in case of error.
} {
    {object_id:integer 0}
    {debug_p 0}
    {node ""}
}

set root_file_id $object_id
if {"" ne $node && [string is integer $node]} { set root_file_id $node }

# --------------------------------------------
# Security & Permissions
#
# set current_user_id [auth::require_login]
set current_user_id 624

set ttt {
    im_object_permissions $current_user_id $object_id view read write admin
    if {!$read} {
	im_rest_error -format "json" -http_status 403 -message "You (user #$current_user_id) have no permissions to read object #$object_id"
	ad_script_abort
    }
}


# --------------------------------------------
# Determine folder for object and create if necessary
#
set folder_id [im_rest_fs_folder_for_object -object_id $object_id]

# permission::require_permission -party_id $current_user_id -object_id $folder_id -privilege "read"

# set package_and_root [fs::get_folder_package_and_root $folder_id]
# set package_id [lindex $package_and_root 0]
# set root_folder_id [lindex $package_and_root 1]
set folder_path ""


# --------------------------------------------
# Main hierarchical SQL
#
set top_level [db_string top_level "select tree_level(tree_sortkey) from cr_items where item_id = :folder_id"]

set sql "
	select
		ci.item_id as file_id,
		ci.live_revision as live_version_id,
		ci.parent_id,
		cr.revision_id,				-- CR revision (OpenACS and PO files)
		fv.version_id,				-- CR revision (only PO files)
		ci.name as file_name,
		tree_level(ci.tree_sortkey) - :top_level as level,
		(select count(*) from cr_items child where child.parent_id = ci.item_id) as num_children,
		CASE	WHEN ci.content_type::text = 'content_folder' THEN 'folder'
			WHEN ci.content_type::text = 'content_extlink' THEN 'url'
			WHEN ci.content_type::text = 'content_symlink' THEN 'symlink'
			WHEN ci.content_type::text = 'file_storage_object' THEN 'file'
			ELSE ci.content_type
		END AS type,
		CASE WHEN bts.open_p = 'o' THEN 'true' ELSE 'false' END as expanded,
		ci.*,
		cr.*,
		fv.*
	from	
		cr_items parent_ci,
		cr_items ci
		LEFT JOIN cr_extlinks ce ON ci.item_id = ce.extlink_id
		LEFT JOIN cr_folders cf ON ci.item_id = cf.folder_id
		JOIN acs_objects o ON ci.item_id = o.object_id
		LEFT OUTER JOIN im_biz_object_tree_status bts ON (
			ci.item_id = bts.object_id and 
			bts.page_url = 'default' and
			bts.user_id = :current_user_id
		)
		LEFT OUTER JOIN im_files f ON (ci.item_id = f.file_id)
		-- add cr_revisions = file_versions using OUTER JOIN
		LEFT OUTER JOIN cr_revisions cr ON (cr.item_id = ci.item_id)
		LEFT OUTER JOIN im_file_versions fv ON (cr.revision_id = fv.version_id)
	where
		parent_ci.item_id = :folder_id and
		ci.tree_sortkey between parent_ci.tree_sortkey and tree_right(parent_ci.tree_sortkey)
		and tree_level(ci.tree_sortkey) > :top_level
	order by
		ci.tree_sortkey,
		cr.revision_id
"
# ad_return_complaint 1 [im_ad_hoc_query -subtotals_p 0 -format html $sql]


# Write the list of files to a multirow,
# while storing the versions in a hash
multirow create file_multirow file_id parent_id revision_id version_id file_name level num_children type expanded
set file_id ""
set old_file_id ""
db_foreach files $sql {

    # Collect file versions in a hash
    if {"" eq $version_status_id} { set version_status_id [im_file_status_active] }
    if {"" eq $version_type_id} { set version_type_id [im_file_type_default] }
    if {"" eq $revision_id} { set revision_id 0 }
    if {"" eq $content_length} { set content_length 0 }
    set version_list [list]
    if {[info exists version_hash($file_id)]} { set version_list $version_hash($file_id) }
    lappend version_list "
            {
                \"version_id\": $revision_id,
                \"content_length\": $content_length,
                \"mime_type\": \"$mime_type\",
                \"title\": \"[im_quotejson $title]\",
                \"description\": \"[im_quotejson $description]\",
                \"status_id\": $version_status_id,
                \"type_id\": $version_type_id,
                \"creation_date\": \"$version_creation_date\",
                \"modification_date\": \"$version_modification_date\",
                \"sha1\": \"$version_sha1\",
                \"upload_comment\": \"[im_quotejson $version_upload_comment]\"
            }"

    set version_hash($file_id) $version_list

    # write out the multirow entry for the file
    if {$old_file_id ne $file_id} {
	multirow append file_multirow $file_id $parent_id $revision_id $version_id $file_name $level $num_children $type $expanded
	set old_file_id $file_id
    }
}

# Write out the last file multirow
if {"" ne $file_id} {
    multirow append file_multirow $file_id $parent_id $revision_id $version_id $file_name $level $num_children $type $expanded
}
#ad_return_complaint 1 [array get version_hash]


# --------------------------------------------
# Get all the extra variables for a file ci_item:
# fraber 2018-08-22: No meta-variables yet...
set valid_vars [list level type]
set valid_vars [lsort -unique $valid_vars]

# needed to calculate sha1
set storage_area_key "CR_FILES"
set storage_area_path [cr_fs_path $storage_area_key]
ns_log Debug "file-tree.json.tcl: storage_area_key($storage_area_key) storage_area_path($storage_area_path)"

set file_json ""
set ctr 0
set old_level 0
set indent ""
template::multirow foreach file_multirow {

    ns_log Notice "file-tree.json.tcl: file_id=$file_id, file_id=$file_id, version_id=$version_id, name($file_name)"
    if {$debug_p} { append file_json "\n// finish: ctr=$ctr, level=$level, old_level=$old_level\n" }
    
    # -----------------------------------------
    # Close off the previous entry
    # -----------------------------------------
    
    # This is the first child of the previous item
    # Increasing the level always happens in steps of 1
    if {$level > $old_level} {
	append file_json ",\n${indent}\t\"children\":\[\n"
    }
    
    # A group of children needs to be closed.
    # Please note that this can cascade down to several levels.
    while {$level < $old_level} {
	append file_json "\n${indent}\}\]\n"
	incr old_level -1
	set indent ""
	for {set i 0} {$i < $old_level} {incr i} { append indent "\t" }
    }
    
    # The current file is on the same level as the previous.
    # This is also executed after reducing the old_level in the previous while loop
    if {$level == $old_level} {
	if {0 != $ctr} { 
	    append file_json "${indent}\n${indent}\},\n"
	}
    }
    
    if {$debug_p} { append file_json "\n// $file_name: ctr=$ctr, level=$level, old_level=$old_level\n" }
    
    set indent ""
    for {set i 0} {$i < $level} {incr i} { append indent "\t" }
    
    if {0 == $num_children} { set leaf_json "true" } else { set leaf_json "false" }
    if {"folder" == $type} { set leaf_json "false" }
    
    set quoted_char_map {"\n" "\\n" "\r" "\\r" "\"" "\\\"" "\\" "\\\\"}
    set quoted_file_name [string map $quoted_char_map $file_name]
    
    # Get the list of versions from hash
    set version_list {}
    if {[info exists version_hash($file_id)]} { set version_list $version_hash($file_id) }
    
    append file_json "${indent}\{
${indent}\t\"id\":$file_id,
${indent}\t\"live_version_id\":$live_version_id,
${indent}\t\"name\":\"$quoted_file_name\",
${indent}\t\"iconCls\":\"icon-$type\",
${indent}\t\"expanded\":$expanded,
${indent}\t\"versions\":\[[join $version_list ", "]
${indent}\t\],
"

    if {"folder" == $type && 0 == $num_children} {
	append file_json "${indent}\t\"children\":\[\],\n"
    }
    
    foreach var $valid_vars {
	# Skip xml_* variables (only used by MS-File)
	if {[regexp {^xml_} $var match]} { continue }
	
	# Append the value to the JSON output
	set value [set $var]
	set quoted_value [string map $quoted_char_map $value]
	append file_json "${indent}\t\"$var\":\"$quoted_value\",\n"
    }
    append file_json "${indent}\t\"leaf\":$leaf_json"
    
    incr ctr
    set old_level $level
}

set level 0
while {$level < $old_level} {
    # A group of children needs to be closed.
    # Please note that this can cascade down to several levels.
    append file_json "\n${indent}\}\]\n"
    incr old_level -1
    set indent ""
    for {set i 0} {$i < $old_level} {incr i} { append indent "\t" }
}
